package utils

import (
	"os"
	"path/filepath"
	"time"
)

// FileWatcher is an utility that
type FileWatcher struct {
	path string

	closeCh          chan struct{}
	pollInterval     time.Duration
	writeGracePeriod time.Duration
	timer            *time.Timer
}

func NewFileWatcher(path string) *FileWatcher {
	return &FileWatcher{
		path:             path,
		writeGracePeriod: 200 * time.Millisecond,
		pollInterval:     1 * time.Second,
		closeCh:          make(chan struct{}),
	}
}

func (fw *FileWatcher) Start(changeHandler func()) {
	prevModTime := fw.fileModTime(fw.path)

	// use polling-based approach to detect file changes
	// we can't use fsnotify/fsnotify due to issues with symlink+socket. see #462.
	go func() {
		for {
			select {
			case _, ok := <-fw.closeCh:
				if !ok {
					return // watcher is closed.
				}
			case <-time.After(fw.pollInterval):
			}

			modTime := fw.fileModTime(fw.path)
			if modTime.IsZero() {
				continue // file does not exists
			}

			if prevModTime != modTime {
				fw.debounce(changeHandler)
				prevModTime = modTime
			}
		}

	}()
}

func (fw *FileWatcher) fileModTime(path string) time.Time {
	// Ensure that the target that we're watching is not a symlink as we won't get any events when we're watching
	// a symlink.
	fileRealPath, err := filepath.EvalSymlinks(path)
	if err != nil {
		return time.Time{}
	}

	info, err := os.Stat(fileRealPath)
	if err != nil {
		return time.Time{}
	}

	return info.ModTime()
}

func (fw *FileWatcher) debounce(fn func()) {
	if fw.timer != nil {
		fw.timer.Stop()
	}

	fw.timer = time.AfterFunc(fw.writeGracePeriod, fn)
}

func (fw *FileWatcher) Stop() {
	close(fw.closeCh)
}
